from collections import defaultdict
from functools import reduce
from pathlib import Path

import tomlkit
import tomlkit.items
from packaging.requirements import Requirement

import ragna
from ragna.core import Assistant, SourceStorage

HERE = Path(__file__).parent
PYPROJECT_TOML = HERE / ".." / "pyproject.toml"


def main():
    optional_dependencies = make_optional_dependencies(
        extract_builtin_document_handler_requirements(),
        extract_builtin_component_requirements(),
    )
    update_pyproject_toml(optional_dependencies)


def make_optional_dependencies(*optional_requirements):
    optional_dependencies = defaultdict(list)
    for requirements in optional_requirements:
        for name, specifiers in requirements.items():
            optional_dependencies[name].extend(specifiers)

    return sorted(
        str(Requirement(f"{name} {reduce(lambda a, b: a & b, specifiers)}"))
        for name, specifiers in optional_dependencies.items()
    )


def extract_builtin_document_handler_requirements():
    requirements = defaultdict(list)
    for obj in ragna.core.__dict__.values():
        if (
            isinstance(obj, type) and issubclass(obj, ragna.core.DocumentHandler)
        ) and obj is not ragna.core.DocumentHandler:
            append_version_specifiers(requirements, obj)

    return dict(requirements)


def extract_builtin_component_requirements():
    requirements = defaultdict(list)
    for module, cls in [
        (ragna.source_storages, SourceStorage),
        (ragna.assistants, Assistant),
    ]:
        for obj in module.__dict__.values():
            if isinstance(obj, type) and issubclass(obj, cls):
                append_version_specifiers(requirements, obj)

    return dict(requirements)


def update_pyproject_toml(optional_dependencies):
    with open(PYPROJECT_TOML) as file:
        document = tomlkit.load(file)

    document["project"]["optional-dependencies"]["all"] = tomlkit.items.Array(
        list(map(tomlkit.items.String.from_raw, optional_dependencies)),
        trivia=tomlkit.items.Trivia(),
        multiline=True,
    )

    with open(PYPROJECT_TOML, "w") as file:
        tomlkit.dump(document, file)


def append_version_specifiers(version_specifiers, obj):
    for requirement in obj.requirements():
        if not isinstance(requirement, ragna.core.PackageRequirement):
            continue

        requirement = requirement._requirement
        version_specifiers[requirement.name].append(requirement.specifier)


if __name__ == "__main__":
    main()



import contextlib
import functools
import inspect
import platform
import shutil
import socket
import subprocess
import sys
import threading
import time

import httpx
import pytest
import redis


@contextlib.contextmanager
def background_subprocess(*args, stdout=sys.stdout, stderr=sys.stdout, **kwargs):
    process = subprocess.Popen(*args, stdout=stdout, stderr=stderr, **kwargs)
    try:
        yield process
    finally:
        process.kill()
        process.communicate()


def timeout_after(seconds=30, *, message=None):
    timeout = f"Timeout after {seconds:.1f} seconds"
    message = timeout if message is None else f"{timeout}: {message}"

    def decorator(fn):
        if is_debugging():
            return fn

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            result = TimeoutError(message)

            def target():
                nonlocal result
                try:
                    result = fn(*args, **kwargs)
                except Exception as exc:
                    result = exc

            thread = threading.Thread(target=target)
            thread.daemon = True

            thread.start()
            thread.join(seconds)

            if isinstance(result, Exception):
                raise result

            return result

        return wrapper

    return decorator


# Vendored from pytest-timeout
# https://github.com/pytest-dev/pytest-timeout/blob/d91e6d8d69ad706e38a2c9de461a72c4d19777ff/pytest_timeout.py#L218-L247
def is_debugging():
    trace_func = sys.gettrace()
    trace_module = None
    if trace_func:
        trace_module = inspect.getmodule(trace_func) or inspect.getmodule(
            trace_func.__class__
        )
    if trace_module:
        parts = trace_module.__name__.split(".")
        for name in {"pydevd", "bdb", "pydevd_frame_evaluator"}:
            if any(part.startswith(name) for part in parts):
                return True
    return False


def get_available_port():
    with socket.socket() as s:
        s.bind(("", 0))
        return s.getsockname()[1]


@contextlib.contextmanager
def redis_server(scheme="redis://"):
    if platform.system() == "Windows":
        raise RuntimeError("redis-server is not available for Windows.")

    port = get_available_port()
    url = f"{scheme}127.0.0.1:{port}"
    redis_server_executable = shutil.which("redis-server")
    if redis_server_executable is None:
        raise RuntimeError("Unable to find redis-server executable")

    with background_subprocess([redis_server_executable, "--port", str(port)]):
        connection = redis.Redis.from_url(url)

        @timeout_after(message=f"Unable to establish connection to {url}")
        def wait_for_redis_server(poll=0.1):
            while True:
                with contextlib.suppress(redis.ConnectionError):
                    if connection.ping():
                        return url

                time.sleep(poll)

        yield wait_for_redis_server()


skip_redis_on_windows = pytest.mark.skipif(
    platform.system() == "Windows",
    reason="redis-server is not available for Windows.",
)


@contextlib.contextmanager
def ragna_worker(config):
    config_path = config.local_cache_root / "ragna.toml"
    if not config_path.exists():
        config.to_file(config_path)

    with background_subprocess(
        [sys.executable, "-m", "ragna", "worker", "--config", str(config_path)]
    ):

        @timeout_after(message="Unable to start worker")
        def wait_for_worker():
            # This seems quite brittle, but I didn't find a better way to check
            # whether the worker is ready. We are checking the logged messages until
            # we see the "ready" message.
            for line in sys.stderr:
                if "Huey consumer started" in line:
                    return

        yield wait_for_worker()


@contextlib.contextmanager
def ragna_api(config, *, start_worker=None):
    config_path = config.local_cache_root / "ragna.toml"
    if not config_path.exists():
        config.to_file(config_path)

    cmd = [sys.executable, "-m", "ragna", "api", "--config", str(config_path)]

    if start_worker is not None:
        cmd.append(f"--{'' if start_worker else 'no-'}start-worker")

    with background_subprocess(cmd):

        @timeout_after(message="Unable to start ragna api")
        def wait_for_ragna_api(poll=0.1):
            url = config.api.url
            while True:
                with contextlib.suppress(httpx.ConnectError):
                    response = httpx.get(url)
                    if response.is_success:
                        return url

                time.sleep(poll)

        yield wait_for_ragna_api()



import asyncio
import contextlib

import pytest

import ragna.core
from ragna import Config, Rag
from ragna.assistants import RagnaDemoAssistant
from ragna.source_storages import RagnaDemoSourceStorage
from tests.utils import ragna_worker, redis_server, skip_redis_on_windows, timeout_after


@pytest.mark.parametrize(
    "queue",
    ["memory", "file_system", pytest.param("redis", marks=skip_redis_on_windows)],
)
def test_e2e(tmp_path, queue):
    if queue == "memory":
        queue_cm = contextlib.nullcontext("memory")
        worker_cm_fn = contextlib.nullcontext
    elif queue == "file_system":
        queue_cm = contextlib.nullcontext(str(tmp_path / "queue"))
        worker_cm_fn = ragna_worker
    elif queue == "redis":
        queue_cm = redis_server()
        worker_cm_fn = ragna_worker

    with queue_cm as queue_url:
        config = Config(
            local_cache_root=tmp_path,
            core=dict(queue_url=queue_url),
        )
        with worker_cm_fn(config):
            check_core(config)


@timeout_after()
def check_core(config):
    document_root = config.local_cache_root / "documents"
    document_root.mkdir()
    document_path = document_root / "test.txt"
    with open(document_path, "w") as file:
        file.write("!\n")

    async def core():
        rag = Rag(config)
        chat = rag.chat(
            documents=[document_path],
            source_storage=RagnaDemoSourceStorage,
            assistant=RagnaDemoAssistant,
        )
        async with chat:
            return await chat.answer("?")

    answer = asyncio.run(core())

    assert isinstance(answer, ragna.core.Message)
    assert answer.role is ragna.core.MessageRole.ASSISTANT
    assert {source.document.name for source in answer.sources} == {document_path.name}



from __future__ import annotations

import datetime
import itertools
import uuid
from typing import Any, Iterable, Optional, Type, TypeVar, Union

import pydantic

from ._components import Assistant, Component, Message, MessageRole, SourceStorage
from ._config import Config
from ._document import Document
from ._queue import Queue
from ._utils import RagnaException, default_user, merge_models

T = TypeVar("T", bound=Component)


class Rag:
    """RAG workflow.

    Args:
        config: Ragna configuration.
        load_components: Whether to load the configured components in the current
            process. If omitted, components will be loaded if a memory queue is
            configured.
    """

    def __init__(
        self,
        config: Optional[Config] = None,
        *,
        load_components: Optional[bool] = None,
    ):
        self.config = config or Config()
        self._queue = Queue(self.config, load_components=load_components)

    def chat(
        self,
        *,
        documents: Iterable[Any],
        source_storage: Union[Type[SourceStorage], SourceStorage, str],
        assistant: Union[Type[Assistant], Assistant, str],
        **params: Any,
    ) -> Chat:
        """Create a new [ragna.core.Chat][].

        Args:
            documents: Documents to use. Items that are not [ragna.core.Document][]s are
                passed as positional argument to the configured document class.

                !!! note

                    The default configuration uses [ragna.core.LocalDocument][] as
                    document class. It accepts a path as positional input to create it.
                    Thus, in this configuration you can pass paths as documents.
            source_storage: Source storage to use. If [str][] can be the
                [ragna.core.Component.display_name][] of any configured source
                storage.
            assistant: Assistant to use. If [str][] can be the
                [ragna.core.Component.display_name][] of any configured assistant.
            **params: Additional parameters passed to the source storage and assistant.
        """
        return Chat(
            self,
            documents=documents,
            source_storage=source_storage,
            assistant=assistant,
            **params,
        )


class SpecialChatParams(pydantic.BaseModel):
    user: str = pydantic.Field(default_factory=default_user)
    chat_id: uuid.UUID = pydantic.Field(default_factory=uuid.uuid4)
    chat_name: str = pydantic.Field(
        default_factory=lambda: f"Chat {datetime.datetime.now():%x %X}"
    )


class Chat:
    """
    !!! tip

        This object is usually not instantiated manually, but rather through
        [ragna.core.Rag.chat][].

    A chat needs to be [`prepare`][ragna.core.Chat.prepare]d before prompts can be
    [`answer`][ragna.core.Chat.answer]ed.

    Can be used as context manager to automatically invoke
    [`prepare`][ragna.core.Chat.prepare]:

    ```python
    rag = Rag()

    async with rag.chat(
        documents=[path],
        source_storage=ragna.core.RagnaDemoSourceStorage,
        assistant=ragna.core.RagnaDemoAssistant,
    ) as chat:
        print(await chat.answer("What is Ragna?"))
    ```

    Args:
        rag: The RAG workflow this chat is associated with.
        documents: Documents to use. Items that are not [ragna.core.Document][]s are
            passed as positional argument to the configured document class.

            !!! note

                The default configuration uses [ragna.core.LocalDocument][] as document
                class. It accepts a path as positional input to create it. Thus, in
                this configuration you can pass paths as documents.
        source_storage: Source storage to use. If [str][] can be the
            [ragna.core.Component.display_name][] of any configured source storage.
        assistant: Assistant to use. If [str][] can be the
            [ragna.core.Component.display_name][] of any configured assistant.
        **params: Additional parameters passed to the source storage and assistant.
    """

    def __init__(
        self,
        rag: Rag,
        *,
        documents: Iterable[Any],
        source_storage: Union[Type[SourceStorage], SourceStorage, str],
        assistant: Union[Type[Assistant], Assistant, str],
        **params: Any,
    ) -> None:
        self._rag = rag

        self.documents = self._parse_documents(documents)
        self.source_storage = self._rag._queue.parse_component(source_storage)
        self.assistant = self._rag._queue.parse_component(assistant)

        special_params = SpecialChatParams().model_dump()
        special_params.update(params)
        params = special_params
        self.params = params
        self._unpacked_params = self._unpack_chat_params(params)

        self._prepared = False
        self._messages: list[Message] = []

    async def prepare(self) -> Message:
        """Prepare the chat.

        This [`store`][ragna.core.SourceStorage.store]s the documents in the selected
        source storage. Afterwards prompts can be [`answer`][ragna.core.Chat.answer]ed.

        Returns:
            Welcome message.

        Raises:
            ragna.core.RagnaException: If chat is already
                [`prepare`][ragna.core.Chat.prepare]d.
        """
        if self._prepared:
            raise RagnaException(
                "Chat is already prepared",
                chat=self,
                http_status_code=400,
                detail=RagnaException.EVENT,
            )

        await self._enqueue(self.source_storage, "store", self.documents)
        self._prepared = True

        welcome = Message(
            content="How can I help you with the documents?",
            role=MessageRole.SYSTEM,
        )
        self._messages.append(welcome)
        return welcome

    async def answer(self, prompt: str) -> Message:
        """Answer a prompt.

        Returns:
            Answer.

        Raises:
            ragna.core.RagnaException: If chat is not
                [`prepare`][ragna.core.Chat.prepare]d.
        """
        if not self._prepared:
            raise RagnaException(
                "Chat is not prepared",
                chat=self,
                http_status_code=400,
                detail=RagnaException.EVENT,
            )

        prompt = Message(content=prompt, role=MessageRole.USER)
        self._messages.append(prompt)

        sources = await self._enqueue(
            self.source_storage, "retrieve", self.documents, prompt.content
        )
        answer = Message(
            content=await self._enqueue(
                self.assistant, "answer", prompt.content, sources
            ),
            role=MessageRole.ASSISTANT,
            sources=sources,
        )
        self._messages.append(answer)

        # FIXME: add error handling
        # return (
        #     "I'm sorry, but I'm having trouble helping you at this time. "
        #     "Please retry later. "
        #     "If this issue persists, please contact your administrator."
        # )

        return answer

    def _parse_documents(self, documents: Iterable[Any]) -> list[Document]:
        documents_ = []
        for document in documents:
            if not isinstance(document, Document):
                document = self._rag.config.core.document(
                    document  # type: ignore[misc, call-arg]
                )

            if not document.is_readable():
                raise RagnaException(
                    "Document not readable",
                    document=document,
                    http_status_code=404,
                )

            documents_.append(document)
        return documents_

    def _unpack_chat_params(
        self, params: dict[str, Any]
    ) -> dict[tuple[Type[Component], str], dict[str, Any]]:
        source_storage_models = self.source_storage._protocol_models()
        assistant_models = self.assistant._protocol_models()

        ChatModel = merge_models(
            str(self.params["chat_id"]),
            SpecialChatParams,
            *source_storage_models.values(),
            *assistant_models.values(),
            config=pydantic.ConfigDict(extra="forbid"),
        )

        chat_params = ChatModel.model_validate(params, strict=True).model_dump(
            exclude_none=True
        )
        return {
            component_and_action: model(**chat_params).model_dump()
            for component_and_action, model in itertools.chain(
                source_storage_models.items(), assistant_models.items()
            )
        }

    async def _enqueue(
        self, component: Type[Component], action: str, *args: Any
    ) -> Any:
        try:
            return await self._rag._queue.enqueue(
                component, action, args, self._unpacked_params[(component, action)]
            )
        except RagnaException as exc:
            exc.extra["component"] = component.display_name()
            exc.extra["action"] = action
            raise exc

    async def __aenter__(self) -> Chat:
        await self.prepare()
        return self

    async def __aexit__(
        self, exc_type: Type[Exception], exc: Exception, traceback: str
    ) -> None:
        pass



"""
UI Helpers
"""
import panel as pn


def divider():
    return pn.layout.Divider(styles={"padding": "0em 1em"})


"""
CSS constants
"""

MAIN_COLOR = "#DF5538"  # "rgba(223, 85, 56, 1)"


# MAIN_COLOR_LIGHT = "#10BBE580"
# MAIN_COLOR_LIGHTER = "#E1E8E3"
# TABS_SIDEBAR_BKGROUND_COLOR = "#EAEAEA"
# TABS_SIDEBAR_WIDTH = "20em"

# set modal height
CONFIG_MODAL_MIN_HEIGHT = 610
CONFIG_MODAL_MAX_HEIGHT = 850
CONFIG_MODAL_WIDTH = 800

WELCOME_MODAL_HEIGHT = 275
WELCOME_MODAL_WIDTH = 530


APP_RAW = """

:root {
    --body-font: 'Inter', sans-serif !important;
    --accent-color: {{MAIN_COLOR}} !important;
}

* {
    font-family: 'Inter', sans-serif;
}

.main {
    padding: 0px !important;

}

.pn-wrapper {
    padding: 0px;
}

div.card-margin {
    margin: 0px !important;
    height: 100% !important;
}


/* Hide the whole header */

#header {
    display: none;
}


div#content {
    height: calc(100vh);
}

/* Fix the size of the modal */
#pn-Modal {
    --dialog-width: 800px !important;
    --dialog-height:500px !important; 
}

/* Hide the default close button of the modal */
.pn-modal-close {
    display: none !important;
}


/* Hide the fullscreen button of the template */ 
span.fullscreen-button {
    display:none;
}

""".replace("{{MAIN_COLOR}}", MAIN_COLOR)


CHAT_INTERFACE_CUSTOM_BUTTON = """
:host(.solid) .bk-btn.bk-btn-default {
    background-color: transparent;
    color: gray;
}

:host {
    transform: translate(14px, -56px); 
}

.bk-btn {
    border-radius: 0;
    padding: 0;
    font-size: 14px;
}
    """

BK_INPUT_GRAY_BORDER = (
    """ .bk-input {border-color: var(--neutral-color) !important;} """
)


SS_MULTI_SELECT_STYLE = """
option:hover, option:checked, option:focus {
    color:white !important;
}
"""

SS_LABEL_STYLE = """
:host {
    margin-top:20px;
}

label, .bk-slider-title {
    font-weight: bold;
    margin-bottom: 5px;
}

.bk-slider-value {
    font-weight: normal;
}

.noUi-target {
    border-color: var(--accent-fill-active);
}

.noUi-handle {
    border-radius: 50%;
}

.noUi-handle, .noUi-connects {
    background-color: var(--clear-button-active);
}

.noUi-target[disabled='true'] {
    border-color: darkgray;
}


.noUi-target[disabled='true'] .noUi-connect {
    background-color: darkgray !important;
}


:host(.disabled) div.bk-slider-title {
    color: darkgray !important;
}


"""


SS_ADVANCED_UI_CARD = """
:host {
    border : none !important;
    outline: none !important;
    background-color: unset !important;
}
"""


message_loading_indicator = f""" 
            <div style="height:48px;">
                <svg xmlns="http://www.w3.org/2000/svg" width="48" height="48" viewBox="0 0 24 24">
                    <circle cx="18" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin=".67" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                    <circle cx="12" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin=".33" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                    <circle cx="6" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin="0" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                </svg>
            </div>
            """



try:
    from ._version import __version__
except ModuleNotFoundError:
    import warnings

    warnings.warn("ragna was not properly installed!")
    del warnings

    __version__ = "UNKNOWN"

from . import assistants, core, source_storages
from .core import Config, Rag



import uuid

from ragna.core import (
    Config,
    Document,
    Source,
)

from ._vector_database import VectorDatabaseSourceStorage


class Chroma(VectorDatabaseSourceStorage):
    """[Chroma vector database](https://www.trychroma.com/)

    !!! info "Required packages"

        - `chromadb>=0.4.13`
    """

    # Note that this class has no extra requirements, since the chromadb package is
    # already required for the base class.

    def __init__(self, config: Config) -> None:
        super().__init__(config)

        import chromadb

        self._client = chromadb.Client(
            chromadb.config.Settings(
                is_persistent=True,
                persist_directory=str(self.config.local_cache_root / "chroma"),
                anonymized_telemetry=False,
            )
        )

    def store(
        self,
        documents: list[Document],
        *,
        chat_id: uuid.UUID,
        chunk_size: int = 500,
        chunk_overlap: int = 250,
    ) -> None:
        collection = self._client.create_collection(
            str(chat_id), embedding_function=self._embedding_function
        )

        ids = []
        texts = []
        metadatas = []
        for document in documents:
            for chunk in self._chunk_pages(
                document.extract_pages(),
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap,
            ):
                ids.append(str(uuid.uuid4()))
                texts.append(chunk.text)
                metadatas.append(
                    {
                        "document_id": str(document.id),
                        "page_numbers": self._page_numbers_to_str(chunk.page_numbers),
                        "num_tokens": chunk.num_tokens,
                    }
                )

        collection.add(
            ids=ids,
            documents=texts,
            metadatas=metadatas,  # type: ignore[arg-type]
        )

    def retrieve(
        self,
        documents: list[Document],
        prompt: str,
        *,
        chat_id: uuid.UUID,
        chunk_size: int = 500,
        num_tokens: int = 1024,
    ) -> list[Source]:
        collection = self._client.get_collection(
            str(chat_id), embedding_function=self._embedding_function
        )

        result = collection.query(
            query_texts=prompt,
            n_results=min(
                # We cannot retrieve source by a maximum number of tokens. Thus, we
                # estimate how many sources we have to query. We overestimate by a
                # factor of two to avoid retrieving to few sources and needed to query
                # again.
                # ---
                # FIXME: querying only a low number of documents can lead to not finding
                #  the most relevant one.
                #  See https://github.com/chroma-core/chroma/issues/1205 for details.
                #  Instead of just querying more documents here, we should use the
                #  appropriate index parameters when creating the collection. However,
                #  they are undocumented for now.
                max(int(num_tokens * 2 / chunk_size), 100),
                collection.count(),
            ),
            include=["distances", "metadatas", "documents"],
        )

        num_results = len(result["ids"][0])
        result = {
            key: [None] * num_results if value is None else value[0]  # type: ignore[index]
            for key, value in result.items()
        }
        # dict of lists -> list of dicts
        results = [
            {key[:-1]: value[idx] for key, value in result.items()}
            for idx in range(num_results)
        ]

        # That should be the default, but let's make extra sure here
        results = sorted(results, key=lambda r: r["distance"])

        # TODO: we should have some functionality here to remove results with a high
        #  distance to keep only "valid" sources. However, there are two issues:
        #  1. A "high distance" is fairly subjective
        #  2. Whatever threshold we use is very much dependent on the encoding method
        #  Thus, we likely need to have a callable parameter for this class

        document_map = {str(document.id): document for document in documents}
        return self._take_sources_up_to_max_tokens(
            (
                Source(
                    id=result["id"],
                    document=document_map[result["metadata"]["document_id"]],
                    location=result["metadata"]["page_numbers"],
                    content=result["document"],
                    num_tokens=result["metadata"]["num_tokens"],
                )
                for result in results
            ),
            max_tokens=num_tokens,
        )



import os
import uuid
from typing import Any

from ragna.core import (
    Config,
    Document,
    EnvVarRequirement,
    PackageRequirement,
    RagnaException,
    Requirement,
)


class S3Document(Document):
    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [
            PackageRequirement("boto3"),
            EnvVarRequirement("AWS_ACCESS_KEY_ID"),
            EnvVarRequirement("AWS_SECRET_ACCESS_KEY"),
            EnvVarRequirement("AWS_REGION"),
            EnvVarRequirement("AWS_S3_BUCKET"),
        ]

    @classmethod
    def _session(cls):
        import boto3

        return boto3.Session(
            aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
            region_name=os.environ["AWS_REGION"],
        )

    @classmethod
    async def get_upload_info(
        cls, *, config: Config, user: str, id: uuid.UUID, name: str
    ) -> tuple[str, dict[str, Any], dict[str, Any]]:
        if not PackageRequirement("boto3").is_available():
            raise RagnaException()

        session = cls._session()
        s3 = session.client("s3")

        bucket = os.environ["AWS_S3_BUCKET"]
        response = s3.generate_presigned_post(
            Bucket=bucket,
            Key=str(id),
            ExpiresIn=config.api.upload_token_ttl,
        )

        url = response["url"]
        data = response["fields"]
        metadata = {"bucket": bucket}

        return url, data, metadata

    def is_readable(self) -> bool:
        session = self._session()
        s3 = session.resource("s3")

        import botocore.exceptions

        try:
            s3.Object(self.metadata["bucket"], str(self.id)).load()
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return False

            raise RagnaException() from error

        return True

    def read(self) -> bytes:
        session = self._session()
        s3 = session.resource("s3")
        return s3.Object(self.metadata["bucket"], str(self.id)).get()["Body"].read()



import itertools
import platform
import re
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast
from urllib.parse import urlsplit

import huey.api
import huey.constants
import huey.contrib.asyncio
import huey.utils
import redis

from ._components import Component
from ._config import Config
from ._utils import RagnaException


def task_config(
    retries: int = 0, retry_delay: int = 0
) -> Callable[[Callable], Callable]:
    def decorator(fn: Callable) -> Callable:
        fn.__ragna_task_config__ = (  # type: ignore[attr-defined]
            dict(retries=retries, retry_delay=retry_delay)
        )
        return fn

    return decorator


_COMPONENTS: dict[Type[Component], Optional[Component]] = {}

R = TypeVar("R")


def execute(
    component: Type[Component],
    fn: Callable[..., R],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> R:
    self = _COMPONENTS[component]
    assert self is not None
    return fn(self, *args, **kwargs)


class _Task(huey.api.Task):
    def execute(self) -> Any:
        return execute(*self.args)


T = TypeVar("T", bound=Component)


class Queue:
    def __init__(self, config: Config, *, load_components: Optional[bool]) -> None:
        self._config = config
        self._huey = self._load_huey(config.core.queue_url)

        for component in itertools.chain(
            config.core.source_storages,
            config.core.assistants,
        ):
            self.parse_component(component, load=load_components)

    def _load_huey(self, url: str) -> huey.Huey:
        # FIXME: we need to store_none=True here. SourceStorage.store returns None and
        #  if we wouldn't store it, waiting for a result is timing out. Maybe there is a
        #  better way to do this?
        common_kwargs = dict(name="ragna", store_none=True)
        if url == "memory":
            _huey = huey.MemoryHuey(immediate=True, **common_kwargs)
        elif platform.system() == "Windows" and re.match(r"\w:\\", url):
            # This special cases absolute paths on Windows, e.g. C:\Users\...,
            # since they don't play well with urlsplit below
            _huey = huey.FileHuey(path=url, use_thread_lock=True, **common_kwargs)
        else:
            components = urlsplit(url)
            if components.scheme in {"", "file"}:
                _huey = huey.FileHuey(
                    path=components.path, use_thread_lock=True, **common_kwargs
                )
            elif components.scheme in {"redis", "rediss"}:
                _huey = huey.RedisHuey(url=url, **common_kwargs)
                try:
                    _huey.storage.conn.ping()
                except redis.exceptions.ConnectionError:
                    raise RagnaException("Unable to connect to redis", url=url)
            else:
                raise RagnaException("Unknown URL scheme", url=url)
        # This is registering the execute function above to be called if a task is
        # enqueued. We need to create the TaskWrapper object here, because this is the
        # only way to dynamically register tasks while staying in the public API. This
        # could be replaced by
        # self._huey._registry._registry[f"{__name__}.{_Task.__name__}"] = _Task
        huey.api.TaskWrapper(_huey, execute, name=_Task.__name__)

        return _huey

    def parse_component(
        self, component: Union[Type[T], T, str], *, load: Optional[bool] = None
    ) -> Type[T]:
        if load is None:
            load = isinstance(self._huey, huey.MemoryHuey)

        if isinstance(component, type) and issubclass(component, Component):
            cls = component
            instance = None
        elif isinstance(component, Component):
            cls = cast(Type[T], type(component))
            instance = component
        elif isinstance(component, str):
            try:
                cls = cast(
                    Type[T],
                    next(cls for cls in _COMPONENTS if cls.display_name() == component),
                )
            except StopIteration:
                raise RagnaException("Unknown component", component=component)
            instance = None

        instance = _COMPONENTS.setdefault(cls, instance)

        if instance is not None:
            return cls

        if load:
            if not cls.is_available():
                raise RagnaException("Component not available", name=cls.display_name())

            instance = cls(self._config)

        _COMPONENTS[cls] = instance

        return cls

    async def enqueue(
        self,
        component: Type[Component],
        action: str,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        fn = getattr(component, action)
        task = _Task(
            args=(component, fn, args, kwargs),
            **getattr(fn, "__ragna_task_config__", dict()),
        )
        result = self._huey.enqueue(task)
        output = await huey.contrib.asyncio.aget_result(result)
        if isinstance(output, huey.utils.Error):
            raise RagnaException("Task failed", **output.metadata)
        return output

    def create_worker(self, num_workers: int = 1) -> huey.api.Consumer:
        return self._huey.create_consumer(
            workers=num_workers, worker_type=huey.constants.WORKER_THREAD
        )



import abc
import os

import ragna
from ragna.core import (
    Assistant,
    Config,
    EnvVarRequirement,
    Requirement,
    Source,
    task_config,
)


class ApiAssistant(Assistant):
    _API_KEY_ENV_VAR: str

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [EnvVarRequirement(cls._API_KEY_ENV_VAR)]

    def __init__(self, config: Config) -> None:
        super().__init__(config)

        import httpx

        self._client = httpx.Client(
            headers={"User-Agent": f"{ragna.__version__}/{self}"},
            timeout=60,
        )
        self._api_key = os.environ[self._API_KEY_ENV_VAR]

    @task_config(retries=2, retry_delay=1)
    def answer(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int = 256
    ) -> str:
        return self._call_api(prompt, sources, max_new_tokens=max_new_tokens)

    @abc.abstractmethod
    def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> str:
        ...



import logging
import subprocess
import sys
import time
from pathlib import Path
from typing import Annotated, Optional
from urllib.parse import urlsplit

import httpx
import rich
import typer
import uvicorn

import ragna
from ragna._api import app as api_app
from ragna._ui import app as ui_app
from ragna._utils import timeout_after
from ragna.core._queue import Queue

from .config import ConfigOption, check_config, init_config

app = typer.Typer(
    name="ragna",
    invoke_without_command=True,
    no_args_is_help=True,
    add_completion=False,
    pretty_exceptions_enable=False,
)


def version_callback(value: bool) -> None:
    if value:
        rich.print(f"ragna {ragna.__version__} from {ragna.__path__[0]}")
        raise typer.Exit()


@app.callback()
def _main(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version", callback=version_callback, help="Show version and exit."
        ),
    ] = None,
) -> None:
    pass


@app.command(help="Start a wizard to build a Ragna configuration interactively.")
def init(
    *,
    output_path: Annotated[
        Path,
        typer.Option(
            "-o",
            "--output-file",
            metavar="OUTPUT_PATH",
            default_factory=lambda: Path.cwd() / "ragna.toml",
            show_default="./ragna.toml",
            help="Write configuration to <OUTPUT_PATH>.",
        ),
    ],
    force: Annotated[
        bool,
        typer.Option(
            "-f", "--force", help="Overwrite an existing file at <OUTPUT_PATH>."
        ),
    ] = False,
) -> None:
    config, output_path, force = init_config(output_path=output_path, force=force)
    config.to_file(output_path, force=force)


@app.command(help="Check the availability of components.")
def check(config: ConfigOption = "./ragna.toml") -> None:  # type: ignore[assignment]
    is_available = check_config(config)
    raise typer.Exit(int(not is_available))


@app.command(help="Start workers.")
def worker(
    *,
    config: ConfigOption = "./ragna.toml",  # type: ignore[assignment]
    num_threads: Annotated[
        int,
        typer.Option("--num-threads", "-n", help="Number of worker threads to start."),
    ] = 1,
) -> None:
    if config.core.queue_url == "memory":
        rich.print(f"With {config.core.queue_url=} no worker is required!")
        raise typer.Exit(1)

    queue = Queue(config, load_components=True)
    worker = queue.create_worker(num_threads)

    # FIXME: we need to configure this properly
    logging.basicConfig(level=logging.INFO)
    worker.run()


@app.command(help="Start the REST API.")
def api(
    *,
    config: ConfigOption = "./ragna.toml",  # type: ignore[assignment]
    start_worker: Annotated[
        Optional[bool],
        typer.Option(
            help="Start a ragna worker alongside the REST API in a subprocess.",
            show_default="Start if a non-memory queue is configured.",
        ),
    ] = None,
) -> None:
    if start_worker is None:
        start_worker = config.core.queue_url != "memory"

    components = urlsplit(config.api.url)
    if components.hostname is None or components.port is None:
        # TODO: make this part of the config validation
        rich.print(f"Unable to extract hostname and port from {config.api.url}.")
        raise typer.Exit(1)

    if start_worker:
        process = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "ragna",
                "worker",
                "--config",
                config.__ragna_cli_config_path__,  # type: ignore[attr-defined]
            ],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
    else:
        process = None

    try:
        uvicorn.run(
            api_app(config), host=components.hostname, port=components.port or 31476
        )
    finally:
        if process is not None:
            process.kill()
            process.communicate()


@app.command(help="Start the UI.")
def ui(
    *,
    config: ConfigOption = "./ragna.toml",  # type: ignore[assignment]
    start_api: Annotated[
        Optional[bool],
        typer.Option(
            help="Start the ragna REST API alongside the UI in a subprocess.",
            show_default="Start if the API is not served at the configured URL.",
        ),
    ] = None,
) -> None:
    def check_api_available() -> bool:
        try:
            httpx.get(config.api.url)
            return True
        except httpx.ConnectError:
            return False

    if start_api is None:
        start_api = not check_api_available()

    if start_api:
        process = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "ragna",
                "api",
                "--config",
                config.__ragna_cli_config_path__,  # type: ignore[attr-defined]
            ],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
    else:
        process = None

    try:
        if process is not None:

            @timeout_after(30)
            def wait_for_api() -> None:
                while not check_api_available():
                    time.sleep(0.5)

            wait_for_api()

        ui_app(config).serve()  # type: ignore[no-untyped-call]
    finally:
        if process is not None:
            process.kill()
            process.communicate()



from __future__ import annotations

from pathlib import Path
from typing import Union

import tomlkit
from pydantic import Field, ImportString, field_validator
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

from ._authentication import Authentication
from ._components import Assistant, SourceStorage
from ._document import Document
from ._utils import RagnaException


class ConfigBase:
    @classmethod
    def customise_sources(
        cls,
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # This order is needed to prioritize values from environment variables over
        # values from a configuration file. However, since the config instantiation from
        # a config file goes through the regular constructor of the Python object, we
        # are also implicitly prioritizing environment variables over values passed
        # explicitly passed to the constructor. For example, if the environment variable
        # 'RAGNA_RAG_DATABASE_URL' is set, any values passed to
        # `RagnaConfig(rag=RagConfig(database_url=...))` is ignored.
        # TODO: Find a way to achieve the following priorities:
        #  1. Explicitly passed to Python object
        #  2. Environment variable
        #  3. Configuration file
        #  4. Default
        return env_settings, init_settings


class CoreConfig(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="ragna_core_")

    queue_url: str = "memory"

    document: ImportString[type[Document]] = "ragna.core.LocalDocument"  # type: ignore[assignment]
    source_storages: list[ImportString[type[SourceStorage]]] = [
        "ragna.source_storages.RagnaDemoSourceStorage"  # type: ignore[list-item]
    ]
    assistants: list[ImportString[type[Assistant]]] = [
        "ragna.assistants.RagnaDemoAssistant"  # type: ignore[list-item]
    ]


class ApiConfig(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="ragna_api_")

    url: str = "http://127.0.0.1:31476"
    # FIXME: this needs to be dynamic for the UI url
    origins: list[str] = ["http://127.0.0.1:31477"]
    database_url: str = "memory"

    authentication: ImportString[
        type[Authentication]
    ] = "ragna.core.RagnaDemoAuthentication"  # type: ignore[assignment]


class UiConfig(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="ragna_ui_")

    url: str = "http://127.0.0.1:31477"
    # FIXME: this needs to be dynamic for the url
    origins: list[str] = ["http://127.0.0.1:31477"]


class Config(BaseSettings):
    """Ragna configuration"""

    model_config = SettingsConfigDict(env_prefix="ragna_")

    local_cache_root: Path = Field(
        default_factory=lambda: Path.home() / ".cache" / "ragna"
    )

    @field_validator("local_cache_root")
    @classmethod
    def _resolve_and_make_path(cls, path: Path) -> Path:
        path = path.expanduser().resolve()
        path.mkdir(parents=True, exist_ok=True)
        return path

    core: CoreConfig = Field(default_factory=CoreConfig)
    api: ApiConfig = Field(default_factory=ApiConfig)
    ui: UiConfig = Field(default_factory=UiConfig)

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> Config:
        path = Path(path).expanduser().resolve()
        if not path.is_file():
            raise RagnaException(f"{path} does not exist.")

        with open(path) as file:
            return cls.model_validate(tomlkit.load(file).unwrap())

    def to_file(self, path: Union[str, Path], *, force: bool = False) -> None:
        path = Path(path).expanduser().resolve()
        if path.exists() and not force:
            raise RagnaException(f"{path} already exist.")

        with open(path, "w") as file:
            tomlkit.dump(self.model_dump(mode="json"), file)



from pathlib import Path

import pytest
from mktestdocs import check_md_file

HERE = Path(__file__).parent
DOCS_DIR = (HERE / ".." / "docs").resolve()


@pytest.mark.parametrize("path", DOCS_DIR.glob("**/*.md"), ids=str)
def test_files_good(mocker, path):
    mocker.patch("builtins.open", mocker.mock_open())

    # FIXME: The REST API tutorial uses await outside sync functions
    # this would work in a Jupyter Notebook, but not in a Python script
    # We'll also need to have the API running to test this properly.
    if path.relative_to(DOCS_DIR).as_posix() == "tutorials/rest-api.md":
        with pytest.raises(SyntaxError, match="'await' outside function"):
            check_md_file(path, memory=True)
    else:
        check_md_file(path, memory=True)



import uuid

import panel as pn
import param
from panel.reactive import ReactiveHTML

from . import styles as ui

# TODO : move all the CSS rules in a dedicated file

chat_entry_stylesheets = [
    """ 
            :host .right, :host .center, :host .chat-entry {
                    width:100% !important;
            }
    """,
    """
            :host div.bk-panel-models-layout-Column:not(.left) { 
                    width:100% !important;
            }
    """,
    """
            :host .message {
                width: calc(100% - 15px);
                box-shadow: unset;
            }
    """,
    """
            :host .chat-entry-user {
                background-color: rgba(243, 243, 243);
                border: 1px solid rgb(238, 238, 238);
                margin-bottom: 20px;
            }
    """,
    # The padding bottom is used to give some space for the copy and source info buttons
    """
            :host .chat-entry-ragna, :host .chat-entry-system,  :host .chat-entry-assistant{
                background-color: white;
                border: 1px solid rgb(234, 234, 234);
                padding-bottom: 30px;
                margin-bottom: 20px;
            }
    """,
    """
            :host .avatar {
                margin-top:0px;
                box-shadow: unset;
            }
    """,
    """ 
            :host .left {
                height: unset !important;
                min-height: unset !important;
            }
    """,
    """ 
            :host .right {
                
            }
    """,
]
pn.chat.ChatMessage._stylesheets = (
    pn.chat.ChatMessage._stylesheets + chat_entry_stylesheets
)

markdown_table_stylesheet = """

            /* Better rendering of the markdown tables */
            table { 
                margin-top:10px;
                margin-bottom:10px;
            }
            """


class RagnaChatCopyIcon(ReactiveHTML):
    title = param.String(default=None, doc="The title of the button ")
    value = param.String(default=None, doc="The text to copy to the clipboard.")

    _template = """
        <div type="button" 
                id="copy-button"
                onclick="${script('copy_to_clipboard')}"
                class="container"
                style="cursor: pointer;"
        >
            <svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-clipboard" width="16" height="16" 
                    viewBox="0 0 24 24" stroke-width="2" stroke="gray" fill="none" stroke-linecap="round" stroke-linejoin="round">
                <path stroke="none" d="M0 0h24v24H0z" fill="none"/>
                <path d="M9 5h-2a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h10a2 2 0 0 0 2 -2v-12a2 2 0 0 0 -2 -2h-2" />
                <path d="M9 3m0 2a2 2 0 0 1 2 -2h2a2 2 0 0 1 2 2v0a2 2 0 0 1 -2 2h-2a2 2 0 0 1 -2 -2z" />
            </svg>
            <span>${title}</span>
        </div>            
        """

    _scripts = {
        "copy_to_clipboard": """navigator.clipboard.writeText(`${data.value}`);"""
    }

    _stylesheets = [
        """ div.container { 
                            
                            display:flex;
                            flex-direction: row;
                            margin: 7px 10px;
                    
                            color:gray;
                        } 
                    
                        svg {
                            margin-right:5px;
                        }

                    """
    ]


class RagnaChatMessage(pn.chat.ChatMessage):
    msg_data = param.Dict(default={})
    on_click_source_info_callback = param.Callable(default=None)

    def __init__(self, msg_data, user, on_click_source_info_callback=None, **kwargs):
        self.role = msg_data["role"]

        params = {
            "msg_data": msg_data,
            # user is the name of the assistant (eg 'Ragna/DemoAssistant')
            # or the name of the user, depending on the role
            "user": user,
            "on_click_source_info_callback": on_click_source_info_callback,
            "object": msg_data["content"],
            "renderers": [
                lambda txt: RagnaChatMessage.chat_entry_value_renderer(
                    txt, role=self.role
                )
            ],
            "show_timestamp": False,
            "show_reaction_icons": False,
            "show_copy_icon": False,
            "show_user": False,
        }

        params["avatar"] = RagnaChatMessage.get_avatar(self.role, user)

        super().__init__(**(params | kwargs))

        self.update_css_classes()
        self.chat_copy_icon.visible = False

        if self.role == "assistant":
            source_info_button = pn.widgets.Button(
                name="Source Info",
                icon="info-circle",
                stylesheets=[
                    ui.CHAT_INTERFACE_CUSTOM_BUTTON,
                ],
            )

            source_info_button.on_click(self.trigger_on_click_source_info_callback)

            copy_button = RagnaChatCopyIcon(
                value=self.object,
                title="Copy",
                stylesheets=[
                    ui.CHAT_INTERFACE_CUSTOM_BUTTON,
                ],
            )

            self._composite[1].append(pn.Row(copy_button, source_info_button, height=0))

    def trigger_on_click_source_info_callback(self, event):
        if self.on_click_source_info_callback is not None:
            self.on_click_source_info_callback(event, self)

    def update_css_classes(self):
        role = self.msg_data["role"] if "role" in self.msg_data else None
        self.css_classes = ["chat-entry", f"chat-entry-{role}"]

    @classmethod
    def get_avatar(cls, role, user) -> str:
        if role == "system":
            return "imgs/ragna_logo.svg"
        elif role == "user":
            # FIXME: user needs to be dynamic based on the username that was logged in with
            return "👤"
        elif role == "assistant":
            # FIXME: This needs to represent the assistant somehow
            if user == "Ragna/DemoAssistant":
                return "imgs/ragna_logo.svg"
            elif user.startswith("OpenAI/gpt-3.5"):
                return pn.chat.message.GPT_3_LOGO
            elif user == "OpenAI/gpt-4":
                return pn.chat.message.GPT_4_LOGO

            return "🤖"

        # should never happen
        return "?"

    @classmethod
    def chat_entry_value_renderer(cls, txt, role):
        markdown_css_classes = []
        if role is not None:
            markdown_css_classes = [
                f"chat-entry-{role}",
            ]

        return pn.pane.Markdown(
            txt,
            css_classes=markdown_css_classes,
            stylesheets=[markdown_table_stylesheet],
        )


class RagnaChatInterface(pn.chat.ChatInterface):
    def __init__(self, *objects, **params):
        super().__init__(*objects, **params)

    @param.depends("placeholder_text", watch=True, on_init=True)
    def _update_placeholder(self):
        loading_avatar = RagnaChatMessage.get_avatar("system", None)

        self._placeholder = RagnaChatMessage(
            {
                "role": "system",
                "content": ui.message_loading_indicator,
            },
            user=" ",
            show_timestamp=False,
            avatar=loading_avatar,
            reaction_icons={},
            show_copy_icon=False,
        )


class CentralView(pn.viewable.Viewer):
    current_chat = param.ClassSelector(class_=dict, default=None)
    trigger_scroll_to_latest = param.Integer(default=0)

    def __init__(self, api_wrapper, **params):
        super().__init__(**params)

        self.api_wrapper = api_wrapper
        self.on_click_chat_info = None

    def on_click_chat_info_wrapper(self, event):
        if self.on_click_chat_info is not None:
            pills = "".join(
                [
                    f"""<div class='chat_document_pill'>{d['name']}</div>"""
                    for d in self.current_chat["metadata"]["documents"]
                ]
            )

            grid_height = len(self.current_chat["metadata"]["documents"]) // 3

            advanced_config_md = "\n".join(
                f"- **{key.replace('_', ' ').title()}**: {value}"
                for key, value in self.current_chat["metadata"]["params"].items()
            )

            markdown = [
                "To change configurations, start a new chat.\n",
                "**Uploaded Files**",
                f"<div class='pills_list'>{pills}</div><br />\n\n",
                "----",
                "**Source Storage**",
                f"""<span>{self.current_chat['metadata']['source_storage']}</span>\n""",
                "----",
                "**Assistant**",
                f"""<span>{self.current_chat['metadata']['assistant']}</span>\n""",
                "**Advanced configuration**",
                advanced_config_md,
            ]

            markdown = "\n".join(markdown)

            self.on_click_chat_info(
                event,
                "Chat Config",
                [
                    pn.pane.Markdown(
                        markdown,
                        dedent=True,
                        # debug
                        # pn.pane.Markdown(f"Chat ID: {self.current_chat['id']}"),
                        stylesheets=[
                            """ :host {  width: 100%; } 
                                                 
                                                .pills_list {
                                                    /*background-color:gold;*/
                                                    display: grid;
                                                    grid-auto-flow: row;
                                                    row-gap: 10px;
                                                    grid-template: repeat({{GRID_HEIGHT}}, 1fr) / repeat(3, 1fr);
                                                    max-height: 200px;
                                                    overflow: scroll;
                                                }
                                                 
                                                .chat_document_pill {
                                                                    background-color: rgb(241,241,241);
                                                                    
                                                                    margin-left: 5px;   
                                                                    margin-right: 5px;
                                                                    padding: 5px 15px;
                                                                    border-radius: 10px;
                                                                    color:var(--accent-color);
                                                                    width:fit-content;
                                                                    grid-column: span 1;
                                                                    
                                                                }   
                                                ul {
                                                    list-style-type: none
                                                }                                                 

                                                """.replace(
                                "{{GRID_HEIGHT}}", str(grid_height)
                            )
                        ],
                    ),
                ],
            )

    def on_click_source_info_wrapper(self, event, msg):
        if self.on_click_chat_info is not None:
            markdown = "This response was generated using the following data from the uploaded files: <br />\n"

            for i in range(len(msg.msg_data["sources"])):
                source = msg.msg_data["sources"][i]

                location = ""
                if source["location"] != "":
                    location = f": {source['location']}"
                markdown += (
                    f"""{(i+1)}. **{source['document']['name']}** {location}\n"""
                )
                markdown += "----\n"

            self.on_click_chat_info(
                event,
                "Source Info",
                [
                    pn.pane.Markdown(
                        markdown,
                        dedent=True,
                        stylesheets=[""" hr { width: 94%; height:1px;  }  """],
                    ),
                ],
            )

    def set_current_chat(self, chat):
        self.current_chat = chat

    async def chat_callback(
        self, contents: str, user: str, instance: pn.chat.ChatInterface
    ):
        self.current_chat["messages"].append({"role": "user", "content": contents})

        try:
            answer = await self.api_wrapper.answer(self.current_chat["id"], contents)

            self.current_chat["messages"].append(answer["message"])

            yield {
                "user": "Ragna",
                "avatar": "🤖",
                "value": answer["message"]["content"],
            }
        except Exception as e:
            print(e)

            yield {
                "user": "Ragna",
                "avatar": RagnaChatMessage.get_avatar("system", None),
                "value": "Sorry, something went wrong. If this problem persists, please contact your administrator.",
            }

    def get_chat_messages(self):
        chat_entries = []

        if self.current_chat is not None:
            assistant = self.current_chat["metadata"]["assistant"]
            # FIXME: user needs to be dynamic based on the username that was logged in with
            username = "User"

            for m in self.current_chat["messages"]:
                chat_entry = RagnaChatMessage(
                    m,
                    username if m["role"] == "user" else assistant,
                    on_click_source_info_callback=self.on_click_source_info_wrapper,
                )
                chat_entries.append(chat_entry)

        return chat_entries

    @pn.depends("current_chat")
    def chat_interface(self):
        if self.current_chat is None:
            return

        chat_interface = RagnaChatInterface(
            callback=self.chat_callback,
            callback_user="Ragna",
            show_rerun=False,
            show_undo=False,
            show_clear=False,
            show_button_name=False,
            view_latest=True,
            sizing_mode="stretch_width",
            auto_send_types=[],
            widgets=[
                pn.widgets.TextInput(
                    placeholder="Ask Ragna...",
                    stylesheets=[
                        """:host input[type="text"] { 
                                                                border:none !important;
                                                                box-shadow: 0px 0px 6px 0px rgba(0, 0, 0, 0.2);
                                                                padding: 10px 10px 10px 15px;
                                                            }
                                                        
                                                            :host input[type="text"]:focus { 
                                                                box-shadow: 0px 0px 8px 0px rgba(0, 0, 0, 0.3);
                                                            }

                                                         """
                    ],
                )
            ],
            renderers=[
                lambda txt: RagnaChatMessage.chat_entry_value_renderer(txt, role=None)
            ],
            message_params={
                "show_reaction_icons": False,
                "show_user": False,
                "show_copy_icon": False,
                "show_timestamp": False,
                # the proper avatar for the assistant is not when replacing the default ChatMessage objects
                # with RagnaChatMessage objects.
                "avatar_lookup": lambda user: "👤" if user == "User" else None,
            },
        )

        chat_interface._card.stylesheets += [
            """ 
                                             
                                            :host { 
                                                border:none !important;
                                            }

                                            .chat-feed-log {  
                                                padding-right: 18%;
                                                margin-left: 18% ;
                                                padding-top:25px !important;
                                                
                                            }
                                             
                                            .chat-interface-input-container {
                                                margin-left:19%;
                                                margin-right:20%;
                                                margin-bottom: 20px;
                                            }


                                            """
        ]

        """
        By default, each new message is a ChatMessage object. 
        But for new messages from the AI, we want to have a RagnaChatMessage, that contains the msg data, the sources, etc.
        I haven't found a better way than to watch for the `objects` param of chat_interface, 
            and replace the ChatMessage objects with RagnaChatMessage object.
        We do it only for the new messages from the rag, not for the existing messages, neither for the messages from the user.
        """

        def messages_changed(event):
            if len(chat_interface.objects) != len(self.current_chat["messages"]):
                return

            assistant = self.current_chat["metadata"]["assistant"]
            # FIXME: user needs to be dynamic based on the username that was logged in with
            username = "User"

            needs_refresh = False
            for i in range(len(chat_interface.objects)):
                msg = chat_interface.objects[i]

                if not isinstance(msg, RagnaChatMessage) and msg.user != "User":
                    chat_interface.objects[i] = RagnaChatMessage(
                        self.current_chat["messages"][i],
                        username
                        if self.current_chat["messages"][i]["role"] == "user"
                        else assistant,
                        on_click_source_info_callback=self.on_click_source_info_wrapper,
                    )
                    msg = chat_interface.objects[i]
                    needs_refresh = True

            if needs_refresh:
                chat_interface._chat_log.param.trigger("objects")

        chat_interface.param.watch(
            messages_changed,
            ["objects"],
        )

        # Here, we build a list of RagnaChatMessages from the existing messages of this chat,
        # and set them as the content of the chat interface
        chat_interface.objects = self.get_chat_messages()

        # Now that setting all the objects is done, we can watch the change of objects,
        # ie new messages being appended to the chat. When that happens,
        # make sure we scroll to the latest msg.
        chat_interface.param.watch(
            lambda event: self.param.trigger("trigger_scroll_to_latest"),
            ["objects"],
        )

        return chat_interface

    @pn.depends("current_chat", "trigger_scroll_to_latest")
    def scroll_to_latest_fix(self):
        """
        This snippet needs to be re-rendered many times so the scroll-to-latest happens:
            - each time the current chat changes, hence the pn.depends on current_chat
            - each time a message is appended to the chat, hence the pn.depends on trigger_scroll_to_latest.
                trigger_scroll_to_latest is triggered in the chat_interface method, when chat_interface.objects changes.

        Twist : the HTML script node needs to have a different ID each time it is rendered,
                otherwise the browser doesn't re-render it / doesn't execute the JS part.
                Hence the random ID.
        """

        random_id = str(uuid.uuid4())

        return pn.pane.HTML(
            """<script id="{{RANDOM_ID}}" type="text/javascript">

                            setTimeout(() => {
                                    var chatbox_scrolldiv = $$$('.chat-feed-log')[0];
                                    chatbox_scrolldiv.scrollTop = chatbox_scrolldiv.scrollHeight;
                            }, 150);
                            
            </script>""".replace("{{RANDOM_ID}}", random_id)
        )

    @pn.depends("current_chat")
    def header(self):
        if self.current_chat is None:
            return

        current_chat_name = ""
        if self.current_chat is not None:
            current_chat_name = self.current_chat["metadata"]["name"]

        chat_name_header = pn.pane.HTML(
            f"<p>{current_chat_name}</p>",
            sizing_mode="stretch_width",
            stylesheets=[
                """ 

                        :host p {
                            max-width: 50%;
                            height:100%;

                            text-overflow: ellipsis;
                            white-space: nowrap;
                            overflow: hidden;
                            margin: 0px 0px 0px 10px;

                            font-size:20px;
                            text-decoration: underline;
                            text-underline-offset: 4px;

                            /* I don't understand why this is necessary to vertically align the text ... */
                            line-height:250%; 
                            
                        }
                        """
            ],
        )

        chat_documents_pills = []
        if (
            self.current_chat is not None
            and "metadata" in self.current_chat
            and "documents" in self.current_chat["metadata"]
        ):
            doc_names = [d["name"] for d in self.current_chat["metadata"]["documents"]]

            # FIXME: Instead of setting a hard limit of 20 documents here, this should
            #  scale automatically with the width of page
            #  See https://github.com/Quansight/ragna/issues/224
            for doc_name in doc_names[:20]:
                pill = pn.pane.HTML(
                    f"""<div class="chat_document_pill">{doc_name}</div>""",
                    stylesheets=[
                        """
                                                 :host {
                                                    background-color: rgb(241,241,241);
                                                    margin-top: 15px;
                                                    margin-left: 5px;   
                                                    margin-right: 5px;
                                                    padding: 5px 15px;
                                                    border-radius: 10px;
                                                    color:var(--accent-color);
                                                    
                                                 }   

                                                 """
                    ],
                )

                chat_documents_pills.append(pill)

        chat_info_button = None
        if self.current_chat is not None:
            button_name = f"{self.current_chat['metadata']['assistant']} | {self.current_chat['metadata']['source_storage']}"
            chat_info_button = pn.widgets.Button(
                name=button_name, button_style="outline", icon="info-circle"
            )
            chat_info_button.stylesheets.append(
                """
                                        :host {  
                                                    margin-top:10px;
                                        }
                                        
                                    """
            )
            chat_info_button.on_click(self.on_click_chat_info_wrapper)

        return pn.Row(
            chat_name_header,
            *chat_documents_pills,
            chat_info_button,
            stylesheets=[
                """:host {  
                                    background-color: #F9F9F9;
                                    border-bottom: 1px solid #EEEEEE;
                                    width: 100% !important;
                                    margin:0px;
                                    height:54px;
                                }

                                :host div {
                                    
                                    vertical-align: middle;
                                }
"""
            ],
        )

    def set_loading(self, is_loading):
        self.main_column.loading = is_loading

    def __panel__(self):
        """
        The ChatInterface.view_latest option doesn't seem to work.
        So to scroll to the latest message, we use some JS trick.

        There might be a more elegant solution than running this after a timeout of 200ms,
        but without it, the $$$ function isn't available yet.
        And even if I add the $$$ here, the fix itself doesn't work and the chat doesn't scroll
        to the bottom.
        """

        self.main_column = pn.Column(
            self.header,
            self.chat_interface,
            self.scroll_to_latest_fix,
            sizing_mode="stretch_width",
            stylesheets=[
                """                    :host { 
                                            background-color: #F9F9F9;
                                            
                                            height:100%;
                                            max-width: 100%;
                                            margin-left: min(15px, 2%);
                                            border-left: 1px solid #EEEEEE;
                                            border-right: 1px solid #EEEEEE;
                                        }
                                """
            ],
        )

        return self.main_column



import panel as pn
import param


class RightSidebar(pn.viewable.Viewer):
    title = param.String(default="")
    content = param.List(default=[])

    def __init__(self, **params):
        super().__init__(**params)

        self.main_column = None
        self.close_button = None

    def show(self):
        self.main_column.css_classes = ["visible_sidebar"]

    def hide(self, event):
        self.main_column.css_classes = ["hidden_sidebar"]

    @pn.depends("content")
    def content_layout(self):
        return pn.Column(*self.content, stylesheets=[""" :host { width: 100%; } """])

    @pn.depends("title")
    def header(self):
        return pn.pane.Markdown(
            f"## {self.title}",
            stylesheets=[
                """ :host { 
                            background-color: rgb(238, 238, 238);
                            margin:0;
                            padding-left:15px !important;
                            width:100%;
                    } """
            ],
        )

    def __panel__(self):
        self.close_button = pn.widgets.Button(
            icon="x",
            button_type="light",
            css_classes=["close_button"],
            stylesheets=[
                """ 
                        :host {
                            position: absolute;
                            top: 6px;
                            right: 10px;
                            z-index: 99;
                        }
                        """
            ],
        )
        self.close_button.on_click(self.hide)

        self.main_column = pn.Column(
            self.close_button,
            self.header,
            self.content_layout,
            stylesheets=[
                """   
                                :host { 
                                        height:100%;
                                        min-width: unset;
                                        width: 0px;
                                        overflow:hidden;

                                        margin-left: min(15px, 2%);
                                        border-left: 1px solid #EEEEEE;
                                }

                                .bk-panel-models-layout-Column {
                                    width: 100%;
                                }

                                :host .close_button {
                                    transform: translateX(20px);

                                }



                                :host(.visible_sidebar) {
                                        animation: 0.25s ease-in forwards show_right_sidebar;
                                }

                                @keyframes show_right_sidebar {
                                    from {
                                        min-width: unset;
                                        width: 0px;
                                    }

                                    to {
                                        min-width: 200px;
                                        width: 25%;
                                    }
                                }

                                
                                :host(.visible_sidebar) .close_button {
                                    animation: 0.25s ease-in forwards show_close_button;
                                }

                                @keyframes show_close_button {
                                    from {
                                        transform: translateX(20px);
                                    }
                                    to {
                                        transform: translateX(0px);
                                    }
                                }

                                /* hide */

                                :host(.hidden_sidebar) {
                                        animation: 0.33s ease-in forwards hide_right_sidebar;
                                }

                                @keyframes hide_right_sidebar {
                                    from {
                                        min-width: 200px;
                                        width: 25%;
                                    }

                                    to {
                                        min-width: unset;
                                        width: 0px;
                                    }
                                }

                                :host(.hidden_sidebar) .close_button {
                                    animation: 0.33s ease-in forwards hide_close_button;
                                }

                                @keyframes hide_close_button {
                                    from {
                                        transform: translateX(0px);
                                    }
                                    to {
                                        transform: translateX(20px);
                                    }
                                }

                                


                                """
            ],
        )

        return self.main_column



import functools
import threading
from typing import Any, Callable
from urllib.parse import SplitResult, urlsplit, urlunsplit


def fix_module(globals: dict[str, Any]) -> None:
    """Fix the __module__ attribute on public objects to hide internal structure.

    Put the following snippet at the end of public `__init__.py` files of ragna
    subpackages. This will hide any internally private structure.

    ```python
    # isort: split

    from ragna._utils import fix_module

    fix_module(globals())
    del fix_module
    ```
    """
    for name, obj in globals.items():
        if name.startswith("_"):
            continue

        obj.__module__ = globals["__package__"]


def _replace_hostname(split_result: SplitResult, hostname: str) -> SplitResult:
    # This is a separate function, since hostname is not an element of the SplitResult
    # namedtuple, but only a property. Thus, we need to replace the netloc item, from
    # which the hostname is generated.
    if split_result.port is None:
        netloc = hostname
    else:
        netloc = f"{hostname}:{split_result.port}"
    return split_result._replace(netloc=netloc)


def handle_localhost_origins(origins: list[str]) -> list[str]:
    # Since localhost is an alias for 127.0.0.1, we allow both so users and developers
    # don't need to worry about it.
    localhost_origins = {
        components.hostname: components
        for url in origins
        if (components := urlsplit(url)).hostname in {"127.0.0.1", "localhost"}
    }
    if "127.0.0.1" in localhost_origins:
        origins.append(
            urlunsplit(_replace_hostname(localhost_origins["127.0.0.1"], "localhost"))
        )
    elif "localhost" in localhost_origins:
        origins.append(
            urlunsplit(_replace_hostname(localhost_origins["localhost"], "127.0.0.1"))
        )

    return origins


def timeout_after(
    seconds: float = 5, *, message: str = ""
) -> Callable[[Callable], Callable]:
    timeout = f"Timeout after {seconds:.1f} seconds"
    message = timeout if message else f"{timeout}: {message}"

    def decorator(fn: Callable) -> Callable:
        @functools.wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            result: Any = TimeoutError(message)

            def target() -> None:
                nonlocal result
                try:
                    result = fn(*args, **kwargs)
                except Exception as exc:
                    result = exc

            thread = threading.Thread(target=target)
            thread.daemon = True

            thread.start()
            thread.join(seconds)

            if isinstance(result, Exception):
                raise result

            return result

        return wrapper

    return decorator
